"""
Command line reference manager with a single source of truth: the .bib file.
Inspired by beets.
"""

import pkg_resources
import sys

import click
import click_plugins
import pybibs
import pyperclip

from . import internals
from . import query

PATH_OPTION = click.Path(writable=True, readable=True, dir_okay=False)
FILE_OPTION = click.option(
    '--file',
    help='File to link to this entry.',
    type=click.Path(exists=True, readable=True, dir_okay=False),
)
DESTINATION_OPTION = click.option(
    '--destination',
    help='A folder to put the file in.',
    type=click.Path(exists=True, readable=True, dir_okay=True, file_okay=False),
)
SEARCH_TERMS_OPTION = click.argument('search_terms', nargs=-1)


@click_plugins.with_plugins(pkg_resources.iter_entry_points('bibo.plugins'))
@click.group(help=__doc__)
@click.option('--database', envvar='BIBO_DATABASE', help='A .bib file.',
              required=True, type=PATH_OPTION)
@click.pass_context
def cli(ctx, database):
    ctx.ensure_object(dict)
    ctx.obj['database'] = database

    # Read the database. Create (in memory) if doesn't exist
    try:
        ctx.obj['data'] = pybibs.read_file(database)
    except FileNotFoundError:
        ctx.obj['data'] = []


@cli.command(short_help='List entries.')
@click.option('--raw', is_flag=True, help='Format as raw .bib entries')
@SEARCH_TERMS_OPTION
@click.pass_context
def list(ctx, search_terms, raw):
    for entry in query.search(ctx.obj['data'], search_terms):
        if raw:
            txt = pybibs.write_string([entry])
        else:
            txt = internals.format_entry(entry)
        click.echo(txt)


@cli.command(short_help='Open the file linked to an entry.')
@SEARCH_TERMS_OPTION
@click.pass_context
def open(ctx, search_terms):
    entry = query.get(ctx.obj['data'], search_terms)

    file_field = entry['fields'].get('file')
    if not file_field:
        click.echo('No file is associated with this entry')
        sys.exit(1)

    internals.open_file(file_field)


@cli.command(short_help='Add a new entry.')
@FILE_OPTION
@DESTINATION_OPTION
@click.pass_context
def add(ctx, destination, **kwargs):
    file_ = kwargs.pop('file')
    file_validation(file_, destination)

    data = ctx.obj['data']
    bib = click.edit(text=pyperclip.paste())
    entry = pybibs.read_entry_string(bib)
    data.append(entry)

    if file_:
        internals.set_file(data, entry, file_, destination)

    pybibs.write_file(data, ctx.obj['database'])


@cli.command(short_help='Remove an entry or a field.')
@SEARCH_TERMS_OPTION
@click.option('--field', help='Field to remove.')
@click.pass_context
def remove(ctx, search_terms, field):
    data = ctx.obj['data']

    entries = [e for e in query.search(data, search_terms)]

    for entry in entries:
        if field is None:
            internals.remove_entry(data, entry)
        else:
            if field in entry['fields']:
                del entry['fields'][field]
            else:
                click.echo('No such field')

    pybibs.write_file(data, ctx.obj['database'])


@cli.command(short_help='Edit an entry.')
@SEARCH_TERMS_OPTION
@click.option('--type', help='Set the type.')
@click.option('--key', help='Set the key.')
@FILE_OPTION
@DESTINATION_OPTION
@click.option('--field', help='Field to edit.')
@click.pass_context
def edit(ctx, search_terms, key, field, destination, **kwargs):
    type_ = kwargs.pop('type')
    file_ = kwargs.pop('file')
    file_validation(file_, destination)

    data = ctx.obj['data']
    entry = query.get(data, search_terms)

    if type_:
        entry['type'] = type_
    if key:
        entry['key'] = key
    if file_:
        internals.set_file(data, entry, file_, destination)
    if field:
        current_value = entry['fields'].get(field, '')
        updated_value = click.edit(text=current_value).strip()
        entry['fields'][field] = updated_value

    pybibs.write_file(data, ctx.obj['database'])


def file_validation(file_, destination):
    if destination and not file_:
        raise click.BadOptionUsage(
            'destination',
            'Specifying destination without a file is meaningless.',
        )

if __name__ == '__main__':
    cli()



import collections
import itertools
import os
import shutil
import subprocess
import sys

import click


def format_entry(entry):
    header_line = header(entry)
    citation_line = citation(entry)
    if citation_line:
        return '\n'.join([header_line, citation_line])
    else:
        return header_line


def header(entry):
    parts = [click.style(entry['key'], fg='green')]
    fields = entry['fields']
    if fields.get('tags'):
        parts.append(click.style(fields['tags'], fg='cyan'))
    if fields.get('file'):
        parts.append('📁')
    if fields.get('url'):
        parts.append('🔗')
    return ' '.join(parts)


def citation(entry):
    fields = entry['fields']
    if 'author' in fields and 'year' in fields and 'title' in fields:
        return '{author} ({year}). {title}'.format_map(fields)
    if 'author' in fields and 'year' in fields:
        return '{author} ({year})'.format_map(fields)
    if 'author' in fields and 'title' in fields:
        return '{author}. {title}'.format_map(fields)
    if 'title' in fields and 'year' in fields:
        return '{title} ({year})'.format_map(fields)
    return None


def open_file(filepath):
    """
    Open file with the default system app.
    Copied from https://stackoverflow.com/a/435669/1224456
    """
    if sys.platform.startswith('darwin'):
        subprocess.Popen(('open', filepath))
    elif os.name == 'nt':
        os.startfile(filepath)
    elif os.name == 'posix':
        subprocess.Popen(('xdg-open', filepath))


def destination_heuristic(data):
    """
    A heuristic to get the folder with all other files from bib, using majority
    vote.
    """
    counter = collections.Counter()
    for entry in data:
        file_field = entry['fields'].get('file')
        if not file_field:
            continue
        path = os.path.dirname(file_field)
        counter[path] += 1

    if not counter:  # No paths found
        raise click.ClickException(
            'Path finding heuristics failed: no paths in the database'
        )

    # Find the paths that appears most often
    sorted_paths = sorted(counter, reverse=True)
    groupby = itertools.groupby(sorted_paths, key=len)
    _, group = next(groupby)

    # We know that there's at least one candidate. Make sure it's
    # the only one
    candidate = next(group)
    try:
        next(group)
    except StopIteration:
        return candidate
    else:
        raise click.ClickException(
            'Path finding heuristics failed: '
            'there are multiple equally valid paths in the database'
        )


def remove_entry(data, entry):
    '''
    Remove an entry in place.
    '''
    file_field = entry['fields'].get('file')
    if file_field:
        try:
            os.remove(file_field)
        except FileNotFoundError:
            click.echo('This entry\'s file was missing')

    data.remove(entry)


def set_file(data, entry, file_, destination=None):
    if not destination:
        destination = destination_heuristic(data)
    _, file_extension = os.path.splitext(file_)
    path = os.path.join(destination, entry['key'] + file_extension)
    entry['fields']['file'] = path
    shutil.copy(file_, path)



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='bibo',
    version='0.0.9',
    description='A reference manager with single source of truth: the .bib file. Inspired by beeets',
    long_description=long_description,
    url='https://github.com/Nagasaki45/bibo',
    author='Tom Gurion',
    author_email='nagasaki45@gmail.com',

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Database :: Front-Ends',
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'Topic :: Text Processing :: Markup :: LaTeX',
        'Topic :: Utilities',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='bib latex reference interface',

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(),

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'click',
        'click-plugins',
        'pybibs',
        'pyperclip',
    ],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    entry_points={
        'console_scripts': [
            'bibo=bibo.bibo:cli',
        ],
    },
)



import itertools

import click


def search(data, search_terms):
    if isinstance(search_terms, str):
        search_terms = [search_terms]
    search_terms = iter(search_terms)
    return _recursive_search(data, search_terms)


def _recursive_search(data, search_terms):
    try:
        search_term = next(search_terms)
        return search(
            (e for e in data if _is_matching(e, search_term)),
            search_terms,
        )
    except StopIteration:
        return data


def _is_matching(entry, search_term):
    search_field, search_value = _parse_search_term(search_term)
    if search_field == 'general':
        return _is_matching_general_field(entry, search_value)
    else:
        return _is_matching_specific_field(entry, search_field, search_value)


def _is_matching_general_field(entry, search_value):
    searchables = itertools.chain(
        entry['fields'].values(),
        (entry[x] for x in ['key', 'type']),
    )

    for searchable in searchables:
        if search_value in searchable.lower():
            return True
    return False


def _is_matching_specific_field(entry, search_field, search_value):
    if search_field == 'key':
        return search_value in entry['key'].lower()
    if search_field == 'type':
        return search_value in entry['type'].lower()
    for field, value in entry['fields'].items():
        if search_field == field.lower():
            return search_value in value.lower()
    return False


def _parse_search_term(search_term):
    parts = search_term.split(':')
    if len(parts) == 1:
        return 'general', parts[0]
    if len(parts) == 2:
        return parts[0].lower(), parts[1].lower()
    raise click.ClickException('Invalid search term')


def get(data, search_terms):
    search_results = search(data, search_terms)

    try:
        entry = next(search_results)
    except StopIteration:
        raise click.ClickException('No entries found')

    try:
        next(search_results)
    except StopIteration:
        pass
    else:
        raise click.ClickException('Multiple entries found')

    return entry
